package ij.plugin.filter;
import java.awt.*;
import java.util.Vector;
import java.util.Properties;
import ij.*;
import ij.gui.*;
import ij.process.*;
import ij.measure.Measurements;

/** This plug-in implements ImageJ's Analyze/Measure and Analyze/Set Measurements commands. */
public class Analyzer implements PlugInFilter, Measurements {
	
	private String arg;
	private ImagePlus imp;
	private int precision = 3;
	
	private static int[] list = {AREA,MEAN,STD_DEV,MODE,MIN_MAX,
		MIN_MAX,CENTROID,CENTROID,CENTER_OF_MASS,CENTER_OF_MASS,PERIMETER,LIMIT};
	private static String[] labels = {"Area","Mean","StdDev","Mode","Min",
		"Max","X","Y","XM","YM","Perim."};
	private static final int UNDEFINED=0,AREAS=1,LENGTHS=2,ANGLES=3,MARK_AND_COUNT=4;
	private static int mode = UNDEFINED;
	private static final String MEASUREMENTS = "measurements";
	private static final String MARK_WIDTH = "mark.width";
	private static int counter;
	private static boolean unsavedMeasurements;
	public static Color darkBlue = new Color(0,0,160);
	private static int measurements = Prefs.getInt(MEASUREMENTS,AREA+MEAN+MIN_MAX);
	public static int markWidth = Prefs.getInt(MARK_WIDTH,3);

	public Analyzer() {
	}
	
	public Analyzer(ImagePlus imp) {
		this.imp = imp;
	}
	
	public int setup(String arg, ImagePlus imp) {
		this.arg = arg;
		this.imp = imp;
		IJ.register(Analyzer.class);
		if (arg.equals("set")) {
			setMeasurements();
			return DONE;
		} else
			return DOES_ALL;
	}

	public void run(ImageProcessor ip) {
		measure();
	}

	void setMeasurements() {
		GenericDialog gd = new GenericDialog("Set Measurements", IJ.getInstance());
		gd.addCheckbox("Area", (measurements&AREA)!=0);
		gd.addCheckbox("Mean Gray Value", (measurements&MEAN)!=0);
		gd.addCheckbox("Standard Deviation", (measurements&STD_DEV)!=0);
		gd.addCheckbox("Modal Gray Value", (measurements&MODE)!=0);
		gd.addCheckbox("Min & Max Gray Value", (measurements&MIN_MAX)!=0);
		gd.addCheckbox("Centroid", (measurements&CENTROID)!=0);
		gd.addCheckbox("Center of Mass", (measurements&CENTER_OF_MASS)!=0);
		gd.addCheckbox("Perimeter", (measurements&PERIMETER)!=0);
		gd.addMessage("");
		gd.addCheckbox("Limit to Threshold", (measurements&LIMIT)!=0);
		gd.showDialog();
		if (gd.wasCanceled())
			return;
		setOptions(gd);
	}
	 
	void setOptions(GenericDialog gd) {
		int oldMeasurements = measurements;
		boolean b=false,skip=false;
		for (int i=0; i<list.length; i++) {
			if (!skip) {
				b = gd.getNextBoolean();
				if (list[i]==MIN_MAX||list[i]==CENTROID||list[i]==CENTER_OF_MASS)
					skip = true;
			} else
				skip = false;
			if (b)
				measurements |= list[i];
			else
				measurements &= ~list[i];
		}
		if ((oldMeasurements&(~LIMIT))!=(measurements&(~LIMIT)))
			mode = UNDEFINED;
	}
	
	void measure() {
		if (Toolbar.getToolId()==Toolbar.CROSSHAIR) {
			markAndCount();
			return;
		}
		Roi roi = imp.getRoi();
		if (roi!=null && roi.getType()>=Roi.LINE) {
			//if (roi.getType()==Roi.POLYLINE && ((PolygonRoi)roi).getAngle()>=0.0)
			//	measureAngle(roi);
			//else
			measureLength(roi);
			return;
		}
		if (mode!=AREAS) {
			if (!resetCounter())
				return;
			displayLabels();
			mode = AREAS;
		}
		ImageStatistics stats = imp.getStatistics(measurements, 256);
		displayResults(stats, imp.getRoi());
	}

	void markAndCount() {
		if (imp.getTitle().equals("Colors"))
			return;
		int x=-1, y=-1;
		ImageWindow win = imp.getWindow();
		if (win!=null) {
			Point p = win.getCanvas().getCursorLoc();
			x = p.x;
			y = p.y;
		}
		if (x<0)
			return;
		imp.killRoi();
		if (mode!=MARK_AND_COUNT) {
			if (!resetCounter())
				return;
			IJ.setColumnHeadings(" \tX\tY\tValue");		
			mode = MARK_AND_COUNT;
		}
		incrementCounter();
		ImageProcessor ip = imp.getProcessor();
		double value = ip.getPixelValue(x,y);
		IJ.write(counter+"\t"+x+"\t"+y+"\t"+n(value));
		if (markWidth>0) {
			ip.setLineWidth(markWidth);
			ip.moveTo(x,y);
			ip.lineTo(x,y);
			imp.updateAndDraw();
			ip.setLineWidth(Line.getWidth());
		}
	}
	
	void measureAngle(Roi roi) {
		if (mode!=ANGLES) {
			if (!resetCounter())
				return;
			IJ.setColumnHeadings(" \tangle");		
			mode = ANGLES;
		}
		incrementCounter();
		IJ.write(counter+"\t"+n(((PolygonRoi)roi).getAngle()));
	}
	
	void measureLength(Roi roi) {
		if (mode!=LENGTHS) {
			if (!resetCounter())
				return;
			IJ.setColumnHeadings(" \tlength");		
			mode = LENGTHS;
		}
		incrementCounter();
		IJ.write(counter+"\t"+n(roi.getLength()));
	}
	
	public void displayLabels() {
		String s = " \t";
		for (int i=0; i<labels.length; i++) {
			if ((measurements&list[i])!=0)
				s += labels[i] + "\t";
		}
		IJ.setColumnHeadings(s);		
	}
	
	public void displayResults(ImageStatistics stats, Roi roi) {
		incrementCounter();
		String s = "";
		s += n(counter);
		double pw,ph;
		if (imp!=null && imp.sCalibrated)
			{pw=imp.pixelWidth; ph=imp.pixelHeight;}
		else
			{pw=1.0; ph=1.0;}
		if ((measurements&AREA)!=0) {
			double area = stats.pixelCount;
			area *= pw*ph; 
			s += n(area);
		}
		if ((measurements&MEAN)!=0) s+=n(stats.mean);
		if ((measurements&STD_DEV)!=0) s+=n(stats.stdDev);
		if ((measurements&MODE)!=0) s+=n(stats.mode);
		if ((measurements&MIN_MAX)!=0) {
			s+=n(stats.min);
			s+=n(stats.max);
		}
		if ((measurements&CENTROID)!=0) {
			s+=n(stats.xCentroid*pw);
			s+=n(stats.yCentroid*ph);
		}
		if ((measurements&CENTER_OF_MASS)!=0) {
			s+=n(stats.xCenterOfMass*pw);
			s+=n(stats.yCenterOfMass*ph);
		}
		if ((measurements&PERIMETER)!=0) {
			double perimeter;
			if (roi!=null)
				perimeter = roi.getLength();
			else
				perimeter = 0.0;
			s+=n(perimeter);
		}
		IJ.write(s);		
	}

	/** Converts a number to a formatted string with a tab at the end. */
	public String n(double n) {
		String s;
		if (Math.round(n)==n)
			s = IJ.d2s(n,0);
		else
			s = IJ.d2s(n,precision);
		return s+"\t";
	}
		
	void incrementCounter() {
		counter++;
		unsavedMeasurements = true;
	}
	
	public static boolean resetCounter() {
		int lineCount = IJ.getTextPanel().getLineCount();
		if (counter>0 && lineCount>0 && unsavedMeasurements) {
			boolean okay = IJ.showMessageWithCancel("Delete Measurements?", 
				"There are "+counter+" measurements that have not\n"
				+"been saved or copied to the clipboard. They\n"
				+"will be deleted if you click OK.");
			if (!okay)
				return false;
		}
		counter = 0;
		return true;
	}
	
	public static void setSaved() {
		unsavedMeasurements = false;
	}
	
	public static int getMeasurements() {
		return measurements;
	}

	public static void savePreferences(Properties prefs) {
		prefs.put(MEASUREMENTS, Integer.toString(measurements));
		prefs.put(MARK_WIDTH, Integer.toString(markWidth));
	}

}
	
